#include <stdio.h>
#include <stdlib.h>		// malloc
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#ifdef MINGW
#include <windows.h>
#include <wincrypt.h>
#endif
#include "aescrypt.h"

/*
 *  encrypt_stream
 */
int
encrypt_mem( u8 * inp, u8 * end, u8 * outp, char *passwd )
{
  aes_context   aes_ctx;
  sha256_context sha_ctx;
  aescrypt_hdr  aeshdr;
  sha256_t      digest;
  u8            exppasswd[160];
  int           passlen;
  u8            IV[16];
  u8            iv_key[48];
  int           i, j, n;
  u8            buffer[32];
  u8            ipad[64], opad[64];
  u8           *outstart = outp;

  passlen = strlen( passwd );
  for ( i = 0; i < passlen; i++ )
  {
    exppasswd[2 * i] = passwd[i];
    exppasswd[2 * i + 1] = 0;
  }
  passlen *= 2;

  // Create the 16-octet IV and 32-octet encryption key
  // used for encrypting the plaintext file.  We do
  // not trust the rand() function, so we improve on
  // that by also hashing the random digits and using
  // only a portion of the hash.  This IV and key
  // generation could be replaced with any good random
  // source of data.
  memset( iv_key, 0, 48 );
  for ( i = 0; i < 48; i += 16 )
  {
    memset( buffer, 0, 32 );
    sha256_starts( &sha_ctx );
    for ( j = 0; j < 256; j++ )
    {
#ifdef MINGW
      CryptGenRandom( random, 32, buffer );
#else
      fread( buffer, 1, 32, random );
#endif
      sha256_update( &sha_ctx, buffer, 32 );
    }
    sha256_finish( &sha_ctx, digest );
    memcpy( iv_key + i, digest, 16 );
  }

  // Write an AES signature at the head of the file, along
  // with the AES file format version number.
  outp += sprintf( ( char * ) outp, "AES" );
  *outp++ = 0x02;		// Version 2
  *outp++ = '\0';		// reserved in version 2
  n = strlen( "DESIGNED BY Acer Incorporated. 1.00" );
  *outp++ = n >> 8;
  *outp++ = n & 0xff;
  outp += sprintf( ( char * ) outp, "DESIGNED BY Acer Incorporated. 1.00" );
  *outp++ = '\0';
  *outp++ = 1;			// extension length
  // Write out the "container" extension
  *outp++ = '\0';
  // Write out 0x0000 to indicate that no more extensions exist
  *outp++ = '\0';
  *outp++ = '\0';

  sha256_starts( &sha_ctx );
  for ( i = 0; i < 256; i++ )
  {
#ifdef MINGW
    CryptGenRandom( random, 32, buffer );
#else
    fread( buffer, 1, 32, random );
#endif
    sha256_update( &sha_ctx, buffer, 32 );
  }
  sha256_finish( &sha_ctx, digest );

  memcpy( IV, digest, 16 );

  // Write the initialization vector to the file
  memcpy( outp, IV, 16 );
  outp += 16;

  // Hash the IV and password 8192 times
  memset( digest, 0, 32 );
  memcpy( digest, IV, 16 );
  for ( i = 0; i < 8192; i++ )
  {
    sha256_starts( &sha_ctx );
    sha256_update( &sha_ctx, digest, 32 );
    sha256_update( &sha_ctx, exppasswd, passlen );
    sha256_finish( &sha_ctx, digest );
  }

  // Set the AES encryption key
  aes_set_key( &aes_ctx, digest, 256 );

  // Set the ipad and opad arrays with values as
  // per RFC 2104 (HMAC).  HMAC is defined as
  // H(K XOR opad, H(K XOR ipad, text))
  memset( ipad, 0x36, 64 );
  memset( opad, 0x5C, 64 );

  for ( i = 0; i < 32; i++ )
  {
    ipad[i] ^= digest[i];
    opad[i] ^= digest[i];
  }

  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, ipad, 64 );

  // Encrypt the IV and key used to encrypt the plaintext file,
  // writing that encrypted text to the output file.
  for ( i = 0; i < 48; i += 16 )
  {
    // Place the next 16 octets of IV and key buffer into
    // the input buffer.
    memcpy( buffer, iv_key + i, 16 );

    // XOR plain text block with previous encrypted
    // output (i.e., use CBC)
    for ( j = 0; j < 16; j++ )
    {
      buffer[j] ^= IV[j];
    }

    // Encrypt the contents of the buffer
    aes_encrypt( &aes_ctx, buffer, buffer );

    // Concatenate the "text" as we compute the HMAC
    sha256_update( &sha_ctx, buffer, 16 );

    // Write the encrypted block
    memcpy( outp, buffer, 16 );
    outp += 16;
    // Update the IV (CBC mode)
    memcpy( IV, buffer, 16 );
  }

  // Write the HMAC
  sha256_finish( &sha_ctx, digest );
  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, opad, 64 );
  sha256_update( &sha_ctx, digest, 32 );
  sha256_finish( &sha_ctx, digest );
  // Write the encrypted block
  memcpy( outp, digest, 32 );
  outp += 32;
  // Re-load the IV and encryption key with the IV and
  // key to now encrypt the datafile.  Also, reset the HMAC
  // computation.
  memcpy( IV, iv_key, 16 );

  // Set the AES encryption key
  aes_set_key( &aes_ctx, iv_key + 16, 256 );

  // Set the ipad and opad arrays with values as
  // per RFC 2104 (HMAC).  HMAC is defined as
  // H(K XOR opad, H(K XOR ipad, text))
  memset( ipad, 0x36, 64 );
  memset( opad, 0x5C, 64 );

  for ( i = 0; i < 32; i++ )
  {
    ipad[i] ^= iv_key[i + 16];
    opad[i] ^= iv_key[i + 16];
  }

  // Wipe the IV and encryption mey from memory
  memset( iv_key, 0, 48 );

  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, ipad, 64 );

  // Initialize the last_block_size value to 0
  aeshdr.last_block_size = 0;

  while ( end > inp )
  {
    n = end - inp;
    if ( n > 16 )
      n = 16;
    memcpy( buffer, inp, n );
    inp += n;
    // XOR plain text block with previous encrypted
    // output (i.e., use CBC)
    for ( i = 0; i < 16; i++ )
      buffer[i] ^= IV[i];

    // Encrypt the contents of the buffer
    aes_encrypt( &aes_ctx, buffer, buffer );

    // Concatenate the "text" as we compute the HMAC
    sha256_update( &sha_ctx, buffer, 16 );

    // Write the encrypted block
    memcpy( outp, buffer, 16 );
    outp += 16;

    // Update the IV (CBC mode)
    memcpy( IV, buffer, 16 );

    // Assume this number of octets is the file modulo
    aeshdr.last_block_size = n;
  }

  // Write the file size modulo
  *outp++ = aeshdr.last_block_size & 0x0F;
  // Write the HMAC
  sha256_finish( &sha_ctx, digest );
  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, opad, 64 );
  sha256_update( &sha_ctx, digest, 32 );
  sha256_finish( &sha_ctx, digest );
  memcpy( outp, digest, 32 );
  outp += 32;
  return outp - outstart;
}

/*
 *  decrypt_stream
 */
int
decrypt_mem( u8 * inp, u8 * end, u8 * outp, char *passwd )
{
  aes_context   aes_ctx;
  sha256_context sha_ctx;
  aescrypt_hdr  aeshdr;
  sha256_t      digest;
  u8            exppasswd[160];
  int           passlen;
  u8            IV[16];
  u8            iv_key[48];
  int           i, j, n, bytes_read;
  u8            buffer[64], buffer2[32];
  u8           *head, *tail;
  u8            ipad[64], opad[64];
  int           reached_eof = 0;
  int           outlen = 0;

  passlen = strlen( passwd );
  for ( i = 0; i < passlen; i++ )
  {
    exppasswd[2 * i] = passwd[i];
    exppasswd[2 * i + 1] = 0;
  }
  passlen *= 2;
  // Read the file header
  memcpy( &aeshdr, inp, sizeof( aescrypt_hdr ) );
  inp += sizeof( aescrypt_hdr );
  if ( !( aeshdr.aes[0] == 'A' && aeshdr.aes[1] == 'E' &&
      aeshdr.aes[2] == 'S' ) )
  {
    fprintf( stderr,
	"Error: Bad file header (not aescrypt file or is corrupted? [%x, %x, %x])\n",
	aeshdr.aes[0], aeshdr.aes[1], aeshdr.aes[2] );
    return -1;
  }

  // Validate the version number and take any version-specific actions
  if ( aeshdr.version == 0 )
  {
    // Let's just consider the least significant nibble to determine
    // the size of the last block
    aeshdr.last_block_size = ( aeshdr.last_block_size & 0x0F );
  }
  else if ( aeshdr.version > 0x02 )
  {
    fprintf( stderr, "Error: Unsupported AES file version: %d\n",
	aeshdr.version );
    return -1;
  }

  // Skip over extensions present v2 and later files
  if ( aeshdr.version >= 0x02 )
  {
    do
    {
      // Determine the extension length, zero means no more extensions
      n = *(short * )inp;
      n = (short)(( n >> 8 ) | ( n << 8 ));	// swap nibbles
      inp += n + sizeof( short);
    }
    while ( n );
  }

  // Read the initialization vector from the file
  memcpy( IV, inp, 16 );
  inp += 16;
  // Hash the IV and password 8192 times
  memset( digest, 0, 32 );
  memcpy( digest, IV, 16 );
  for ( i = 0; i < 8192; i++ )
  {
    sha256_starts( &sha_ctx );
    sha256_update( &sha_ctx, digest, 32 );
    sha256_update( &sha_ctx, exppasswd, passlen );
    sha256_finish( &sha_ctx, digest );
  }

  // Set the AES encryption key
  aes_set_key( &aes_ctx, digest, 256 );

  // Set the ipad and opad arrays with values as
  // per RFC 2104 (HMAC).  HMAC is defined as
  // H(K XOR opad, H(K XOR ipad, text))
  memset( ipad, 0x36, 64 );
  memset( opad, 0x5C, 64 );

  for ( i = 0; i < 32; i++ )
  {
    ipad[i] ^= digest[i];
    opad[i] ^= digest[i];
  }

  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, ipad, 64 );

  // If this is a version 1 or later file, then read the IV and key
  // for decrypting the bulk of the file.
  if ( aeshdr.version >= 0x01 )
  {
    for ( i = 0; i < 48; i += 16 )
    {
      memcpy( buffer, inp, 16 );
      memcpy( buffer2, inp, 16 );
      inp += 16;

      sha256_update( &sha_ctx, buffer, 16 );
      aes_decrypt( &aes_ctx, buffer, buffer );

      // XOR plain text block with previous encrypted
      // output (i.e., use CBC)
      for ( j = 0; j < 16; j++ )
      {
	iv_key[i + j] = ( buffer[j] ^ IV[j] );
      }

      // Update the IV (CBC mode)
      memcpy( IV, buffer2, 16 );
    }

    // Verify that the HMAC is correct
    sha256_finish( &sha_ctx, digest );
    sha256_starts( &sha_ctx );
    sha256_update( &sha_ctx, opad, 64 );
    sha256_update( &sha_ctx, digest, 32 );
    sha256_finish( &sha_ctx, digest );

    memcpy( buffer, inp, 32 );
    inp += 32;
    if ( memcmp( digest, buffer, 32 ) )
      return -1;
    // Re-load the IV and encryption key with the IV and
    // key to now encrypt the datafile.  Also, reset the HMAC
    // computation.
    memcpy( IV, iv_key, 16 );

    // Set the AES encryption key
    aes_set_key( &aes_ctx, iv_key + 16, 256 );

    // Set the ipad and opad arrays with values as
    // per RFC 2104 (HMAC).  HMAC is defined as
    // H(K XOR opad, H(K XOR ipad, text))
    memset( ipad, 0x36, 64 );
    memset( opad, 0x5C, 64 );

    for ( i = 0; i < 32; i++ )
    {
      ipad[i] ^= iv_key[i + 16];
      opad[i] ^= iv_key[i + 16];
    }

    // Wipe the IV and encryption mey from memory
    memset( iv_key, 0, 48 );

    sha256_starts( &sha_ctx );
    sha256_update( &sha_ctx, ipad, 64 );
  }

  // Decrypt the balance of the file

  // Attempt to initialize the ring buffer with contents from the file.
  // Attempt to read 48 octets of the file into the ring buffer.
  bytes_read = end - inp;
  if ( bytes_read > 48 )
    bytes_read = 48;
  memcpy( buffer, inp, bytes_read );
  inp += bytes_read;
  if ( bytes_read < 48 )
  {
    // If there are less than 48 octets, the only valid count
    // is 32 for version 0 (HMAC) and 33 for version 1 or
    // greater files ( file size modulo + HMAC)
    if ( ( aeshdr.version == 0x00 && bytes_read != 32 ) ||
	( aeshdr.version >= 0x01 && bytes_read != 33 ) )
    {
      fprintf( stderr, "Error: Input file is corrupt (1:%d).\n", bytes_read );
      return -1;
    }
    else
    {
      // Version 0 files would have the last block size
      // read as part of the header, so let's grab that
      // value now for version 1 files.
      if ( aeshdr.version >= 0x01 )
      {
	// The first octet must be the indicator of the
	// last block size.
	aeshdr.last_block_size = ( buffer[0] & 0x0F );
      }
      // If this initial read indicates there is no encrypted
      // data, then there should be 0 in the last_block_size field
      if ( aeshdr.last_block_size != 0 )
      {
	fprintf( stderr, "Error: Input file is corrupt (2).\n" );
	return -1;
      }
    }
    reached_eof = 1;
  }
  head = buffer + 48;
  tail = buffer;

  while ( !reached_eof )
  {
    // Check to see if the head of the buffer is past the ring buffer
    if ( head == ( buffer + 64 ) )
      head = buffer;

    bytes_read = end - inp;
    if ( bytes_read > 16 )
      bytes_read = 16;
    memcpy( head, inp, bytes_read );
    inp += bytes_read;
    if ( bytes_read < 16 )
    {
      // The last block for v0 must be 16 and for v1 it must be 1
      if ( ( aeshdr.version == 0x00 && bytes_read > 0 ) ||
	  ( aeshdr.version >= 0x01 && bytes_read != 1 ) )
      {
	fprintf( stderr, "Error: Input file is corrupt (3:%d).\n",
	    bytes_read );
	return -1;
      }

      // If this is a v1 file, then the file modulo is located
      // in the ring buffer at tail + 16 (with consideration
      // given to wrapping around the ring, in which case
      // it would be at buffer[0])
      if ( aeshdr.version >= 0x01 )
      {
	if ( ( tail + 16 ) < ( buffer + 64 ) )
	  aeshdr.last_block_size = ( tail[16] & 0x0F );
	else
	  aeshdr.last_block_size = ( buffer[0] & 0x0F );
      }

      // Indicate that we've reached the end of the file
      reached_eof = 1;
    }

    // Process data that has been read.  Note that if the last
    // read operation returned no additional data, there is still
    // one one ciphertext block for us to process if this is a v0 file.
    if ( ( bytes_read > 0 ) || ( aeshdr.version == 0x00 ) )
    {
      // Advance the head of the buffer forward
      if ( bytes_read > 0 )
	head += 16;

      memcpy( buffer2, tail, 16 );

      sha256_update( &sha_ctx, tail, 16 );
      aes_decrypt( &aes_ctx, tail, tail );

      // XOR plain text block with previous encrypted
      // output (i.e., use CBC)
      for ( i = 0; i < 16; i++ )
	tail[i] ^= IV[i];

      // Update the IV (CBC mode)
      memcpy( IV, buffer2, 16 );

      // If this is the final block, then we may
      // write less than 16 octets
      n = ( ( !reached_eof ) ||
	  ( aeshdr.last_block_size == 0 ) ) ? 16 : aeshdr.last_block_size;

      // Write the decrypted block
      memcpy( outp, tail, n );
      outp += n;
      outlen += n;
      // Move the tail of the ring buffer forward
      tail += 16;
      if ( tail == ( buffer + 64 ) )
	tail = buffer;
    }
  }

  // Verify that the HMAC is correct
  sha256_finish( &sha_ctx, digest );
  sha256_starts( &sha_ctx );
  sha256_update( &sha_ctx, opad, 64 );
  sha256_update( &sha_ctx, digest, 32 );
  sha256_finish( &sha_ctx, digest );

  // Copy the HMAC read from the file into buffer2
  if ( aeshdr.version == 0x00 )
  {
    memcpy( buffer2, tail, 16 );
    tail += 16;
    if ( tail == ( buffer + 64 ) )
      tail = buffer;
    memcpy( buffer2 + 16, tail, 16 );
  }
  else
  {
    memcpy( buffer2, tail + 1, 15 );
    tail += 16;
    if ( tail == ( buffer + 64 ) )
      tail = buffer;
    memcpy( buffer2 + 15, tail, 16 );
    tail += 16;
    if ( tail == ( buffer + 64 ) )
      tail = buffer;
    memcpy( buffer2 + 31, tail, 1 );
  }

  if ( memcmp( digest, buffer2, 32 ) )
    return -1;

  return outlen;
}
