#include <stdio.h>
#include <string.h>
#include <unistd.h>		// getopt
#include <stdlib.h>		// malloc
#include <sys/stat.h>
#ifdef MINGW
#include <windows.h>
#include <wincrypt.h>
#endif
#include "aescrypt.h"
#include "md5.h"

#ifdef MINGW
HCRYPTPROV    random;
#else
FILE         *random;
#endif

#define BUFSZ     512
#define MAX_FILES 80

u8            buf[BUFSZ];

struct file_entry
{
  char          name[128];
  int           len;
  int           offset;
  int           v1;
  int           v2;
};

struct
{
  int           EncryptedSize;
  int           MergeToolVersion;
  int           EncryptionType;
  int           FileCount;
} mhdr;

struct
{
  int           create:1;
  int           overwrite:1;
  int           extract:1;
  int           vfysum:1;
} sw;

char         *passwords[] = {
  "gemuThA8ejAmERu5EGaF9stUprA7r29TadRusWepR5bup64aHecrEfe2e7Eb4dru",
  "FREJE4rEsPAx7naz9T6a3EbrapHEDeFUCHatHenaf99uXU5upastUSEzumEvuwEq",
  "gup2aswuyupadrAKudejepUbruweyaspaspevUzABExEyuFreGuqaRef6uYahesw",
  "merebRu4HUspakEbeXeJa5E432eFeyUbRuYuRu3ubRu4eStEtreS66spa7hefras",
  "tredRaWrudrUthup76rasWUyEV3bRavUbacRUstUSWeqUsWUbRAdruZecreTEdRA",
  "DrevafEbupredreVuc4BubRa8haxazaYUJAt7AkaQEPheC39FrU6RabAtH8WRuTR",
  "THawUCafE6a57uDu9P7chu696SuBreTh4KEcefR7mU333eQaCrAwRab6bABrAkaH",
  "r7qakaqeswazeS9estestatrubRukuqutra44v39u55yecrubreqeqa8wafreche",
  "BR7Mep5UQajUtrE825h89h2crekaNexepres74yuW9asw2pruF6Bru44HafradRa",
  "Nu3raFu3efregu4ufe7at783eprefr2heTr7k3cruRaPhu4A68unat4esavedraV",
  "9ebuw4R6gUxeW5chuSpAde54mEsatrEMAsATaSwUseVenan27hat2awa4eTrusut",
  "PhUkasePePraxutuzebRAStEdameBreSpuQu2eQesTevETRunEdruxUsWuthuRub",
  "wu6RUmapubefepenespaxedrabest2fruvEphespa4radAhechuxagaprareD5th",
  "prE5e2a57aQanEbesa5HufreneDreqaXestebruveDr4yu5uzuxu2ut9cheqafRa",
  "SakAphANUPRAdasTutejaTrutrEstaVECrEbuwuBE5pewaSweyuQUDadatHeJast",
  "C5ejUz3kuyEtrameMeSwAcE6rUBuCreSWu7Repe4rEphaphUpruket5uHemawrEQ",
};

#define NR_PASS (sizeof( passwords ) / sizeof( char * ))

int
create_container( FILE * f, struct file_entry *files )
{
  MD5_CTX       md5;
  FILE         *fcomp;
  int           i, len, offset, key_index, plain_hdrlen, enc_hdrlen;
  u8           *enc_header;
  char         *p;
  struct file_entry *fe;

  // Open the source for random data.  Note that while the entropy
  // might be lower with /dev/urandom than /dev/random, it will not
  // fail to produce something.
#ifdef MINGW
  if ( !CryptAcquireContextW( &random, 0, 0,
      PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT ) )
  {
    fprintf( stderr, "Error acquiring crypt context\n" );
    return 0;
  }
  CryptGenRandom( random, sizeof( int ), ( u8 * ) & key_index );
#else
  if ( ( random = fopen( "/dev/urandom", "r" ) ) == NULL )
  {
    perror( "Error open /dev/urandom:" );
    return 0;
  }
  fread( &key_index, sizeof( int ), 1, random );
#endif
  if ( ( fcomp = fopen( files->name, "rb" ) ) == NULL )
  {
    fprintf( stderr, "Component '%s' not found\n", files->name );
    return 0;
  }
  fread( ( u8 * ) files + files->offset, 1, files->len, fcomp );
  fclose( fcomp );
  if ( strchr( ( char * ) files + files->offset, '\r' ) )
  {
    fprintf( stderr, "'%s' should not contain any carrige returns\n"
	"Use a dos2unix convertor first!\n", files->name );
    return 0;
  }
  plain_hdrlen = files->offset + files->len;
  // extra space (512) for header, IV and checksums
  enc_header = malloc( plain_hdrlen + 512 );
  // choose a random key
  key_index %= NR_PASS;
  enc_hdrlen = encrypt_mem( ( u8 * ) files, ( u8 * ) files + plain_hdrlen,
      enc_header, passwords[key_index] );
  fprintf( f, "%s", "[mgfl_k]:t@ac.tw" );
  // encrypt header just to know how much space to reserve
  enc_hdrlen = encrypt_mem( ( u8 * ) files, ( u8 * ) files + plain_hdrlen,
      enc_header, passwords[key_index] );
  fseek( f, 0x20 + enc_hdrlen, SEEK_SET );	// start of binary files
  for ( fe = files, i = 0; i < mhdr.FileCount; i++, fe++ )
  {
    if ( i )			// append all files except the parameter file
    {
      offset = ftell( f );	// remember the start offset in file
      if ( ( fcomp = fopen( fe->name, "rb" ) ) == NULL )
      {
	fprintf( stderr, "Component '%s' not found\n", fe->name );
	return 0;
      }
      while ( ( len = fread( buf, 1, BUFSZ, fcomp ) ) > 0 )
	fwrite( buf, 1, len, f );
      fclose( fcomp );
      fe->offset = offset;
    }
    // strip the name entry form their path
#ifdef MINGW
    if ( ( p = strrchr( fe->name, '\\' ) ) != NULL )
#else
    if ( ( p = strrchr( fe->name, '/' ) ) != NULL )
#endif
    {
      p++;
      len = strlen( p );
      memmove( fe->name, p, len + 1 );	// include the end zero
    }
    printf( "%10d\t%s\n", fe->len, fe->name );
  }
  // encrypt header again with the correct offsets and names
  enc_hdrlen = encrypt_mem( ( u8 * ) files, ( u8 * ) files + plain_hdrlen,
      enc_header, passwords[key_index] );
  mhdr.EncryptedSize = enc_hdrlen;
  fseek( f, 0x10, SEEK_SET );	// back over to header start
  fwrite( &mhdr, sizeof( int ), 4, f );
  fwrite( enc_header, 1, enc_hdrlen, f );
  free( enc_header );
  // now calculate the proper md5sum of the whole container
  fseek( f, 0, SEEK_SET );	// back at the beginning
  MD5Init( &md5 );
  while ( ( len = fread( buf, 1, BUFSZ, f ) ) > 0 )
    MD5Update( &md5, buf, len );
  MD5Final( &md5 );
  fseek( f, 0, SEEK_SET );	// back at the beginning
  fwrite( md5.digest, 1, 16, f );
  // close random data source
  if ( random != 0 )
  {
#ifdef MINGW
    CryptReleaseContext( random, 0 );
#else
    fclose( random );
#endif
  }
  return 1;
}

int
extract_container( FILE * f )
{
  MD5_CTX       md5;
  char         *md5seed = "[mgfl_k]:t@ac.tw";
  u8            md5sum[16];
  u8            md5buf[64];
  FILE         *fo;
  u8           *bi, *bo;
  int           i, blen, hdlen;
  struct file_entry *fe;

  fread( md5sum, 1, 16, f );
  fread( &mhdr, sizeof( int ), 4, f );
  // sanity check
  if ( mhdr.EncryptedSize > 0xffff || mhdr.FileCount > 500 )
  {
    fprintf( stderr, "Input file doesn't look like an Acer container\n" );
    return 0;
  }
  if ( sw.vfysum )
  {
    // verify checksum 
    if ( fseek( f, 0x10, SEEK_SET ) == -1 )
      return 0;
    MD5Init( &md5 );
    MD5Update( &md5, ( u8 * ) md5seed, 16 );
    while ( ( i = fread( md5buf, 1, 64, f ) ) > 0 )
      MD5Update( &md5, md5buf, i );
    MD5Final( &md5 );
    if ( memcmp( md5.digest, md5sum, 16 ) )
    {
      fprintf( stderr, "Header checksum doesn't match\n" );
      return 0;
    }
  }
  bi = malloc( mhdr.EncryptedSize );
  bo = malloc( mhdr.EncryptedSize );
  if ( fseek( f, 0x20, SEEK_SET ) == -1 )
    return 0;
  fread( bi, 1, mhdr.EncryptedSize, f );
  for ( i = 0; i < NR_PASS; i++ )
  {
    if ( ( hdlen =
	decrypt_mem( bi, bi + mhdr.EncryptedSize, bo, passwords[i] ) ) != -1 )
      break;
  }
  *( bo + hdlen ) = 0;		// string termination
  // for each file entry
  fe = ( struct file_entry * ) bo;
  for ( i = 0; i < mhdr.FileCount; i++, fe++ )
  {
    if ( !strcmp( fe->name, "parameter_list.txt" ) )
    {
      if ( sw.extract )
      {
	fo = fopen( fe->name, "w" );
	fprintf( fo, ( const char * ) bo + fe->offset );
	fclose( fo );
      }
      else
	// use string as format to get DOS CRLF conversion
	printf( ( const char * ) bo + fe->offset );
    }
    else
    {
      printf( "%10d\t%s\n", fe->len, fe->name );
      if ( sw.extract )
      {
	if ( fseek( f, fe->offset, SEEK_SET ) == -1 )
	  return 0;
	// create and copy each component file
	fo = fopen( fe->name, "wb" );
	hdlen = fe->len;
	while ( hdlen > 0 )
	{
	  blen = hdlen < BUFSZ ? hdlen : BUFSZ;
	  blen = fread( buf, 1, blen, f );
	  if ( blen <= 0 )
	    return 0;
	  fwrite( buf, 1, blen, fo );
	  hdlen -= blen;
	}
	fclose( fo );
      }
    }
  }
  return 1;
}

void
usage( char *progname )
{
  char         *p;

#ifdef MINGW  
  p = strrchr( progname, '\\' );
#else  
  p = strrchr( progname, '/' );
#endif  
  if ( p == NULL )
    p = progname;
  else
    p++;
  progname = p;
  fprintf( stderr,
      "Usage: %s [-x[v] | -c[o]] bin_container [component_files]\n",
      progname );
  fprintf( stderr, "\tparameter file may be omitted (zelea2 - June 2013)\n" );
  fprintf( stderr, "\t-x\textract component files (otherwise list only)\n" );
  fprintf( stderr, "\t-v\tverify the container md5sum\n" );
  fprintf( stderr, "\t-c\tcreate container (merged firmware)\n" );
  fprintf( stderr, "\t-o\toverwrite exiting container\n" );
  exit( 0 );
}

int
main( int argc, char *argv[] )
{
  FILE         *f;
  int           i, option, container;

  do
  {
    option = getopt( argc, argv, "hxvco" );
    switch ( option )
    {
      case 'c':
	sw.create = 1;
	break;
      case 'o':
	sw.overwrite = 1;
	break;
      case 'x':
	sw.extract = 1;
	break;
      case 'v':
	sw.vfysum = 1;
	break;
      case 'h':
	usage( argv[0] );
	break;
      case EOF:		// no more options
	break;
      default:
	fprintf( stderr, "getopt returned impossible value: %d ('%c')",
	    option, option );
    }
  }
  while ( option != EOF );
  if ( optind == argc )
    usage( argv[0] );
  container = optind;

  if ( sw.create && sw.extract )
  {
    fprintf( stderr, "Cannot create and extract at the same time!\n\n" );
    usage( argv[0] );
  }
  if ( sw.create )
  {
    struct stat   fstat;
    int           rs = stat( argv[container], &fstat );
    char         *param;
    struct file_entry *files, *fe;

    if ( !sw.overwrite && !rs && fstat.st_size != 0 )	// container exists
    {
      fprintf( stderr,
	  "Container '%s' already exists (use the overwrite switch)\n",
	  argv[container] );
      exit( -1 );
    }
    optind++;
    if ( optind == argc )
    {
      fprintf( stderr, "No components present\n" );
      usage( argv[0] );
    }
    param = "parameter_list.txt";
    if ( strstr( argv[optind], param ) != NULL )
      param = argv[optind++];
    if ( stat( param, &fstat ) || fstat.st_size == 0 )
    {
      fprintf( stderr, "File '%s' doesn't exist or has zero length\n",
	  param );
      exit( -1 );
    }
    mhdr.MergeToolVersion = 4;
    mhdr.EncryptionType = 0;
    mhdr.FileCount = argc - optind + 1;
    files = fe = calloc( mhdr.FileCount * sizeof( struct file_entry ) +
	fstat.st_size + 1, 1 );
    strcpy( fe->name, param );
    fe->offset = mhdr.FileCount * sizeof( struct file_entry );
    fe->len = fstat.st_size;
    fe++;
    for ( i = optind; i < argc; i++, fe++ )
    {
      if ( stat( argv[i], &fstat ) || fstat.st_size == 0 )
      {
	fprintf( stderr, "File '%s' doesn't exist or has zero length\n",
	    argv[i] );
	exit( -1 );
      }
      strcpy( fe->name, argv[i] );
      fe->len = fstat.st_size;
    }
    if ( ( f = fopen( argv[container], "w+b" ) ) == NULL )
    {
      fprintf( stderr, "Cannot create '%s' container\n", argv[container] );
      exit( -1 );
    }
    create_container( f, files );
  }
  else				// list or extract 
  {
    if ( ( f = fopen( argv[optind], "rb" ) ) == NULL )
    {
      fprintf( stderr, "Container '%s' not found\n", argv[optind] );
      exit( -1 );
    }
    if ( extract_container( f ) )
      printf( "\nDone\n" );
  }
  fclose( f );
  exit( 0 );
}
